"""
Support for ipset
"""


import logging

import salt.utils.path
from salt._compat import ipaddress

log = logging.getLogger(__name__)


_IPSET_FAMILIES = {
    "ipv4": "inet",
    "ip4": "inet",
    "ipv6": "inet6",
    "ip6": "inet6",
}

_IPSET_SET_TYPES = {
    "bitmap:ip",
    "bitmap:ip,mac",
    "bitmap:port",
    "hash:ip",
    "hash:mac",
    "hash:ip,port",
    "hash:ip,port,ip",
    "hash:ip,port,net",
    "hash:net",
    "hash:net,net",
    "hash:net,iface",
    "hash:net,port",
    "hash:net,port,net",
    "hash:ip,mark",
    "list:set",
}


_CREATE_OPTIONS = {
    "bitmap:ip": {"range", "netmask", "timeout", "counters", "comment", "skbinfo"},
    "bitmap:ip,mac": {"range", "timeout", "counters", "comment", "skbinfo"},
    "bitmap:port": {"range", "timeout", "counters", "comment", "skbinfo"},
    "hash:ip": {
        "family",
        "hashsize",
        "maxelem",
        "netmask",
        "timeout",
        "counters",
        "comment",
        "skbinfo",
    },
    "hash:mac": {"hashsize", "maxelem", "timeout", "counters", "comment", "skbinfo"},
    "hash:net": {
        "family",
        "hashsize",
        "maxelem",
        "netmask",
        "timeout",
        "counters",
        "comment",
        "skbinfo",
    },
    "hash:net,net": {
        "family",
        "hashsize",
        "maxelem",
        "netmask",
        "timeout",
        "counters",
        "comment",
        "skbinfo",
    },
    "hash:net,port": {
        "family",
        "hashsize",
        "maxelem",
        "netmask",
        "timeout",
        "counters",
        "comment",
        "skbinfo",
    },
    "hash:net,port,net": {
        "family",
        "hashsize",
        "maxelem",
        "netmask",
        "timeout",
        "counters",
        "comment",
        "skbinfo",
    },
    "hash:ip,port,ip": {
        "family",
        "hashsize",
        "maxelem",
        "netmask",
        "timeout",
        "counters",
        "comment",
        "skbinfo",
    },
    "hash:ip,port,net": {
        "family",
        "hashsize",
        "maxelem",
        "netmask",
        "timeout",
        "counters",
        "comment",
        "skbinfo",
    },
    "hash:ip,port": {
        "family",
        "hashsize",
        "maxelem",
        "netmask",
        "timeout",
        "counters",
        "comment",
        "skbinfo",
    },
    "hash:ip,mark": {
        "family",
        "markmask",
        "hashsize",
        "maxelem",
        "timeout",
        "counters",
        "comment",
        "skbinfo",
    },
    "hash:net,iface": {
        "family",
        "hashsize",
        "maxelem",
        "netmask",
        "timeout",
        "counters",
        "comment",
        "skbinfo",
    },
    "list:set": {"size", "timeout", "counters", "comment"},
}

_CREATE_OPTIONS_WITHOUT_VALUE = {"comment", "counters", "skbinfo"}

_CREATE_OPTIONS_REQUIRED = {
    "bitmap:ip": ["range"],
    "bitmap:ip,mac": ["range"],
    "bitmap:port": ["range"],
    "hash:ip": [],
    "hash:mac": [],
    "hash:net": [],
    "hash:net,net": [],
    "hash:ip,port": [],
    "hash:net,port": [],
    "hash:ip,port,ip": [],
    "hash:ip,port,net": [],
    "hash:net,port,net": [],
    "hash:net,iface": [],
    "hash:ip,mark": [],
    "list:set": [],
}


_ADD_OPTIONS = {
    "bitmap:ip": {"timeout", "packets", "bytes", "skbmark", "skbprio", "skbqueue"},
    "bitmap:ip,mac": {"timeout", "packets", "bytes", "skbmark", "skbprio", "skbqueue"},
    "bitmap:port": {"timeout", "packets", "bytes", "skbmark", "skbprio"},
    "hash:ip": {"timeout", "packets", "bytes", "skbmark", "skbprio", "skbqueue"},
    "hash:mac": {"timeout", "packets", "bytes", "skbmark", "skbprio", "skbqueue"},
    "hash:net": {
        "timeout",
        "nomatch",
        "packets",
        "bytes",
        "skbmark",
        "skbprio",
        "skbqueue",
    },
    "hash:net,net": {
        "timeout",
        "nomatch",
        "packets",
        "bytes",
        "skbmark",
        "skbprio",
        "skbqueue",
    },
    "hash:net,port": {
        "timeout",
        "nomatch",
        "packets",
        "bytes",
        "skbmark",
        "skbprio",
        "skbqueue",
    },
    "hash:net,port,net": {
        "timeout",
        "nomatch",
        "packets",
        "bytes",
        "skbmark",
        "skbprio",
        "skbqueue",
    },
    "hash:ip,port,ip": {
        "timeout",
        "packets",
        "bytes",
        "skbmark",
        "skbprio",
        "skbqueue",
    },
    "hash:ip,port,net": {
        "timeout",
        "nomatch",
        "packets",
        "bytes",
        "skbmark",
        "skbprio",
        "skbqueue",
    },
    "hash:ip,port": {
        "timeout",
        "nomatch",
        "packets",
        "bytes",
        "skbmark",
        "skbprio",
        "skbqueue",
    },
    "hash:net,iface": {
        "timeout",
        "nomatch",
        "packets",
        "bytes",
        "skbmark",
        "skbprio",
        "skbqueue",
    },
    "hash:ip,mark": {"timeout", "packets", "bytes", "skbmark", "skbprio", "skbqueue"},
    "list:set": {"timeout", "packets", "bytes", "skbmark", "skbprio", "skbqueue"},
}

__virtualname__ = "ipset"


def __virtual__():
    """
    Only load the module if ipset is installed
    """
    if salt.utils.path.which("ipset"):
        return True
    return (
        False,
        "The ipset execution modules cannot be loaded: ipset binary not in path.",
    )


def _ipset_cmd():
    """
    Return correct command
    """
    return salt.utils.path.which("ipset")


def version():
    """
    Return version from ipset --version

    CLI Example:

    .. code-block:: bash

        salt '*' ipset.version

    """
    cmd = [_ipset_cmd(), "--version"]
    out = __salt__["cmd.run"](cmd, python_shell=False).split()
    return out[1]


def new_set(name=None, set_type=None, family="ipv4", comment=False, **kwargs):
    """
    .. versionadded:: 2014.7.0

    Create new custom set

    CLI Example:

    .. code-block:: bash

        salt '*' ipset.new_set custom_set list:set

        salt '*' ipset.new_set custom_set list:set comment=True

        IPv6:
        salt '*' ipset.new_set custom_set list:set family=ipv6
    """

    ipset_family = _IPSET_FAMILIES[family]
    if not name:
        return "Error: Set Name needs to be specified"

    if not set_type:
        return "Error: Set Type needs to be specified"

    if set_type not in _IPSET_SET_TYPES:
        return "Error: Set Type is invalid"

    # Check for required arguments
    for item in _CREATE_OPTIONS_REQUIRED[set_type]:
        if item not in kwargs:
            return "Error: {} is a required argument".format(item)

    cmd = [_ipset_cmd(), "create", name, set_type]

    for item in _CREATE_OPTIONS[set_type]:
        if item in kwargs:
            if item in _CREATE_OPTIONS_WITHOUT_VALUE:
                cmd.append(item)
            else:
                cmd.extend([item, kwargs[item]])

    # Family only valid for certain set types
    if "family" in _CREATE_OPTIONS[set_type]:
        cmd.extend(["family", cmd, ipset_family])

    if comment:
        cmd.append("comment")

    out = __salt__["cmd.run"](cmd, python_shell=False)

    if not out:
        out = True
    return out


def delete_set(name=None, family="ipv4"):
    """
    .. versionadded:: 2014.7.0

    Delete ipset set.

    CLI Example:

    .. code-block:: bash

        salt '*' ipset.delete_set custom_set

        IPv6:
        salt '*' ipset.delete_set custom_set family=ipv6
    """

    if not name:
        return "Error: Set needs to be specified"

    cmd = [_ipset_cmd(), "destroy", name]
    out = __salt__["cmd.run"](cmd, python_shell=False)

    if not out:
        out = True
    return out


def rename_set(name=None, new_set=None, family="ipv4"):
    """
    .. versionadded:: 2014.7.0

    Delete ipset set.

    CLI Example:

    .. code-block:: bash

        salt '*' ipset.rename_set custom_set new_set=new_set_name

        IPv6:
        salt '*' ipset.rename_set custom_set new_set=new_set_name family=ipv6
    """

    if not name:
        return "Error: Set needs to be specified"

    if not new_set:
        return "Error: New name for set needs to be specified"

    settype = _find_set_type(name)
    if not settype:
        return "Error: Set does not exist"

    settype = _find_set_type(new_set)
    if settype:
        return "Error: New Set already exists"

    cmd = [_ipset_cmd(), "rename", name, new_set]
    out = __salt__["cmd.run"](cmd, python_shell=False)

    if not out:
        out = True
    return out


def list_sets(family="ipv4"):
    """
    .. versionadded:: 2014.7.0

    List all ipset sets.

    CLI Example:

    .. code-block:: bash

        salt '*' ipset.list_sets

    """
    cmd = [_ipset_cmd(), "list", "-t"]
    out = __salt__["cmd.run"](cmd, python_shell=False)

    _tmp = out.split("\n")

    count = 0
    sets = []
    sets.append({})
    for item in _tmp:
        if not item:
            count = count + 1
            sets.append({})
            continue
        key, value = item.split(":", 1)
        sets[count][key] = value[1:]
    return sets


def check_set(name=None, family="ipv4"):
    """
    Check that given ipset set exists.

    .. versionadded:: 2014.7.0

    CLI Example:

    .. code-block:: bash

        salt '*' ipset.check_set name

    """
    if not name:
        return "Error: Set needs to be specified"

    setinfo = _find_set_info(name)
    if not setinfo:
        return False
    return True


def add(name=None, entry=None, family="ipv4", **kwargs):
    """
    Append an entry to the specified set.

    CLI Example:

    .. code-block:: bash

        salt '*' ipset.add name 192.168.1.26

        salt '*' ipset.add name 192.168.0.3,AA:BB:CC:DD:EE:FF

    """
    if not name:
        return "Error: Set needs to be specified"
    if not entry:
        return "Error: Entry needs to be specified"

    setinfo = _find_set_info(name)
    if not setinfo:
        return "Error: Set {} does not exist".format(name)

    settype = setinfo["Type"]

    cmd = [_ipset_cmd(), "add", "-exist", name, entry]

    if "timeout" in kwargs:
        if "timeout" not in setinfo["Header"]:
            return "Error: Set {} not created with timeout support".format(name)

    if "packets" in kwargs or "bytes" in kwargs:
        if "counters" not in setinfo["Header"]:
            return "Error: Set {} not created with counters support".format(name)

    if "comment" in kwargs:
        if "comment" not in setinfo["Header"]:
            return "Error: Set {} not created with comment support".format(name)
        if "comment" not in entry:
            cmd = '{} comment "{}"'.format(cmd, kwargs["comment"])

    if {"skbmark", "skbprio", "skbqueue"} & set(kwargs.keys()):
        if "skbinfo" not in setinfo["Header"]:
            return "Error: Set {} not created with skbinfo support".format(name)

    for item in _ADD_OPTIONS[settype]:
        if item in kwargs:
            cmd.extend([item, kwargs[item]])

    current_members = _find_set_members(name)
    if entry in current_members:
        return "Warn: Entry {} already exists in set {}".format(entry, name)

    # Using -exist to ensure entries are updated if the comment changes
    out = __salt__["cmd.run"](cmd, python_shell=False)

    if not out:
        return "Success"
    return "Error: {}".format(out)


def delete(name=None, entry=None, family="ipv4", **kwargs):
    """
    Delete an entry from the specified set.

    CLI Example:

    .. code-block:: bash

        salt '*' ipset.delete name 192.168.0.3,AA:BB:CC:DD:EE:FF

    """
    if not name:
        return "Error: Set needs to be specified"
    if not entry:
        return "Error: Entry needs to be specified"

    settype = _find_set_type(name)

    if not settype:
        return "Error: Set {} does not exist".format(name)

    cmd = [_ipset_cmd(), "del", name, entry]
    out = __salt__["cmd.run"](cmd, python_shell=False)

    if not out:
        return "Success"
    return "Error: {}".format(out)


def check(name=None, entry=None, family="ipv4"):
    """
    Check that an entry exists in the specified set.

    name
        The ipset name

    entry
        An entry in the ipset.  This parameter can be a single IP address, a
        range of IP addresses, or a subnet block.  Example:

        .. code-block:: cfg

            192.168.0.1
            192.168.0.2-192.168.0.19
            192.168.0.0/25

    family
        IP protocol version: ipv4 or ipv6

    CLI Example:

    .. code-block:: bash

        salt '*' ipset.check name '192.168.0.1 comment "Hello"'

    """
    if not name:
        return "Error: Set needs to be specified"
    if not entry:
        return "Error: Entry needs to be specified"

    settype = _find_set_type(name)
    if not settype:
        return "Error: Set {} does not exist".format(name)

    current_members = _parse_members(settype, _find_set_members(name))

    if not current_members:
        return False

    if isinstance(entry, list):
        entries = _parse_members(settype, entry)
    else:
        entries = [_parse_member(settype, entry)]

    for current_member in current_members:
        for entry in entries:
            if _member_contains(current_member, entry):
                return True

    return False


def test(name=None, entry=None, family="ipv4", **kwargs):
    """
    Test if an entry is in the specified set.

    CLI Example:

    .. code-block:: bash

        salt '*' ipset.test name 192.168.0.2

        IPv6:
        salt '*' ipset.test name fd81:fc56:9ac7::/48
    """
    if not name:
        return "Error: Set needs to be specified"
    if not entry:
        return "Error: Entry needs to be specified"

    settype = _find_set_type(name)
    if not settype:
        return "Error: Set {} does not exist".format(name)

    cmd = [_ipset_cmd(), "test", name, entry]
    out = __salt__["cmd.run_all"](cmd, python_shell=False)

    if out["retcode"] > 0:
        # Entry doesn't exist in set return false
        return False

    return True


def flush(name=None, family="ipv4"):
    """
    Flush entries in the specified set,
    Flush all sets if set is not specified.

    CLI Example:

    .. code-block:: bash

        salt '*' ipset.flush

        salt '*' ipset.flush set

        IPv6:
        salt '*' ipset.flush

        salt '*' ipset.flush set
    """

    cmd = [_ipset_cmd(), "flush"]
    if name:
        cmd.append(name)
    out = __salt__["cmd.run"](cmd, python_shell=False)

    return not out


def _find_set_members(name):
    """
    Return list of members for a set
    """

    cmd = [_ipset_cmd(), "list", name]
    out = __salt__["cmd.run_all"](cmd, python_shell=False)

    if out["retcode"] > 0:
        # Set doesn't exist return false
        return False

    _tmp = out["stdout"].split("\n")
    members = []
    startMembers = False
    for i in _tmp:
        if startMembers:
            members.append(i)
        if "Members:" in i:
            startMembers = True
    return members


def _find_set_info(name):
    """
    Return information about the set
    """

    cmd = [_ipset_cmd(), "list", "-t", name]
    out = __salt__["cmd.run_all"](cmd, python_shell=False)

    if out["retcode"] > 0:
        # Set doesn't exist return false
        return False

    setinfo = {}
    _tmp = out["stdout"].split("\n")
    for item in _tmp:
        # Only split if item has a colon
        if ":" in item:
            key, value = item.split(":", 1)
            setinfo[key] = value[1:]
    return setinfo


def _find_set_type(name):
    """
    Find the type of the set
    """
    setinfo = _find_set_info(name)

    if setinfo:
        return setinfo["Type"]
    else:
        return False


def _parse_members(settype, members):
    if isinstance(members, str):

        return [_parse_member(settype, members)]

    return [_parse_member(settype, member) for member in members]


def _parse_member(settype, member, strict=False):
    subtypes = settype.split(":")[1].split(",")

    all_parts = member.split(" ", 1)
    parts = all_parts[0].split(",")

    parsed_member = []
    for i, subtype in enumerate(subtypes):
        part = parts[i]

        if subtype in ["ip", "net"]:
            try:
                if "/" in part:
                    part = ipaddress.ip_network(part, strict=strict)
                elif "-" in part:
                    start, end = list(map(ipaddress.ip_address, part.split("-")))

                    part = list(ipaddress.summarize_address_range(start, end))
                else:
                    part = ipaddress.ip_address(part)
            except ValueError:
                pass

        elif subtype == "port":
            part = int(part)

        parsed_member.append(part)

    if len(all_parts) > 1:
        parsed_member.append(all_parts[1])

    return parsed_member


def _members_contain(members, entry):
    pass


def _member_contains(member, entry):
    if len(member) < len(entry):
        return False

    for i, _entry in enumerate(entry):
        if not _compare_member_parts(member[i], _entry):
            return False

    return True


def _compare_member_parts(member_part, entry_part):
    if member_part == entry_part:
        # this covers int, string, and equal ip and net
        return True

    # for ip ranges parsed with summarize_address_range
    if isinstance(entry_part, list):
        for entry_part_item in entry_part:
            if not _compare_member_parts(member_part, entry_part_item):
                return False

        return True

    # below we only deal with ip and net objects
    if _is_address(member_part):
        if _is_network(entry_part):
            return member_part in entry_part

    elif _is_network(member_part):
        if _is_address(entry_part):
            return entry_part in member_part

        # both are networks, and == was false
        return False

        # This could be changed to support things like
        # 192.168.0.4/30 contains 192.168.0.4/31
        #
        # return entry_part.network_address in member_part \
        #    and entry_part.broadcast_address in member_part

    return False


def _is_network(o):
    return isinstance(o, (ipaddress.IPv4Network, ipaddress.IPv6Network))


def _is_address(o):
    return isinstance(o, (ipaddress.IPv4Address, ipaddress.IPv6Address))



"""
Simple Smoke Tests for Connected SSH minions
"""

import pytest
from saltfactories.utils.tempfiles import temp_file
from tests.support.case import SSHCase
from tests.support.helpers import requires_system_grains
from tests.support.runtests import RUNTIME_VARS


class SSHMasterTestCase(SSHCase):
    """
    Test ssh master functionality
    """

    @pytest.mark.slow_test
    def test_can_it_ping(self):
        """
        Ensure the proxy can ping
        """
        ret = self.run_function("test.ping")
        self.assertEqual(ret, True)

    @requires_system_grains
    @pytest.mark.slow_test
    @pytest.mark.skip_if_not_root
    def test_service(self, grains):
        service = "cron"
        os_family = grains["os_family"]
        os_release = grains["osrelease"]
        if os_family == "RedHat":
            service = "crond"
        elif os_family == "Arch":
            service = "sshd"
        elif os_family == "MacOS":
            service = "org.ntp.ntpd"
            if int(os_release.split(".")[1]) >= 13:
                service = "com.apple.AirPlayXPCHelper"
        self.run_function("service.enable", [service])
        ret = self.run_function("service.get_all")
        self.assertIn(service, ret)
        self.run_function("service.stop", [service])
        ret = self.run_function("service.status", [service])
        self.assertFalse(ret)
        self.run_function("service.start", [service])
        ret = self.run_function("service.status", [service])
        self.assertTrue(ret)

    @pytest.mark.slow_test
    def test_state_apply(self):
        core_state = """
        {}/testfile:
          file:
            - managed
            - source: salt://testfile
            - makedirs: true
            """.format(
            RUNTIME_VARS.TMP
        )

        with temp_file("core.sls", core_state, RUNTIME_VARS.TMP_BASEENV_STATE_TREE):
            ret = self.run_function("state.apply", ["core"])
            for key, value in ret.items():
                self.assertTrue(value["result"])

    @pytest.mark.slow_test
    def test_state_highstate(self):
        top_sls = """
        base:
          '*':
            - core
            """

        core_state = """
        {}/testfile:
          file:
            - managed
            - source: salt://testfile
            - makedirs: true
            """.format(
            RUNTIME_VARS.TMP
        )

        with temp_file(
            "top.sls", top_sls, RUNTIME_VARS.TMP_BASEENV_STATE_TREE
        ), temp_file("core.sls", core_state, RUNTIME_VARS.TMP_BASEENV_STATE_TREE):
            ret = self.run_function("state.highstate")
            for key, value in ret.items():
                self.assertTrue(value["result"])



import salt.modules.gentoo_service as gentoo_service
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, call, patch
from tests.support.unit import TestCase


class GentooServicesTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.gentoo_service
    """

    def setup_loader_modules(self):
        return {gentoo_service: {}}

    def test_service_list_parser(self):
        """
        Test for parser of rc-status results
        """
        # no services is enabled
        mock = MagicMock(return_value="")
        with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
            self.assertFalse(gentoo_service.get_enabled())
        mock.assert_called_once_with("rc-update -v show")

    def test_get_enabled_single_runlevel(self):
        """
        Test for Return a list of service that are enabled on boot
        """
        service_name = "name"
        runlevels = ["default"]
        mock = MagicMock(return_value=self.__services({service_name: runlevels}))
        with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
            enabled_services = gentoo_service.get_enabled()
            self.assertTrue(service_name in enabled_services)
            self.assertEqual(enabled_services[service_name], runlevels)

    def test_get_enabled_filters_out_disabled_services(self):
        """
        Test for Return a list of service that are enabled on boot
        """
        service_name = "name"
        runlevels = ["default"]
        disabled_service = "disabled"
        service_list = self.__services({service_name: runlevels, disabled_service: []})

        mock = MagicMock(return_value=service_list)
        with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
            enabled_services = gentoo_service.get_enabled()
            self.assertEqual(len(enabled_services), 1)
            self.assertTrue(service_name in enabled_services)
            self.assertEqual(enabled_services[service_name], runlevels)

    def test_get_enabled_with_multiple_runlevels(self):
        """
        Test for Return a list of service that are enabled on boot at more than one runlevel
        """
        service_name = "name"
        runlevels = ["non-default", "default"]
        mock = MagicMock(return_value=self.__services({service_name: runlevels}))
        with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
            enabled_services = gentoo_service.get_enabled()
            self.assertTrue(service_name in enabled_services)
            self.assertEqual(enabled_services[service_name][0], runlevels[1])
            self.assertEqual(enabled_services[service_name][1], runlevels[0])

    def test_get_disabled(self):
        """
        Test for Return a list of service that are installed but disabled
        """
        disabled_service = "disabled"
        enabled_service = "enabled"
        service_list = self.__services(
            {disabled_service: [], enabled_service: ["default"]}
        )
        mock = MagicMock(return_value=service_list)
        with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
            disabled_services = gentoo_service.get_disabled()
            self.assertTrue(len(disabled_services), 1)
            self.assertTrue(disabled_service in disabled_services)

    def test_available(self):
        """
        Test for Returns ``True`` if the specified service is
        available, otherwise returns
        ``False``.
        """
        disabled_service = "disabled"
        enabled_service = "enabled"
        multilevel_service = "multilevel"
        missing_service = "missing"
        shutdown_service = "shutdown"
        service_list = self.__services(
            {
                disabled_service: [],
                enabled_service: ["default"],
                multilevel_service: ["default", "shutdown"],
                shutdown_service: ["shutdown"],
            }
        )
        mock = MagicMock(return_value=service_list)
        with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
            self.assertTrue(gentoo_service.available(enabled_service))
            self.assertTrue(gentoo_service.available(multilevel_service))
            self.assertTrue(gentoo_service.available(disabled_service))
            self.assertTrue(gentoo_service.available(shutdown_service))
            self.assertFalse(gentoo_service.available(missing_service))

    def test_missing(self):
        """
        Test for The inverse of service.available.
        """
        disabled_service = "disabled"
        enabled_service = "enabled"
        service_list = self.__services(
            {disabled_service: [], enabled_service: ["default"]}
        )
        mock = MagicMock(return_value=service_list)
        with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
            self.assertFalse(gentoo_service.missing(enabled_service))
            self.assertFalse(gentoo_service.missing(disabled_service))
            self.assertTrue(gentoo_service.missing("missing"))

    def test_getall(self):
        """
        Test for Return all available boot services
        """
        disabled_service = "disabled"
        enabled_service = "enabled"
        service_list = self.__services(
            {disabled_service: [], enabled_service: ["default"]}
        )
        mock = MagicMock(return_value=service_list)
        with patch.dict(gentoo_service.__salt__, {"cmd.run": mock}):
            all_services = gentoo_service.get_all()
            self.assertEqual(len(all_services), 2)
            self.assertTrue(disabled_service in all_services)
            self.assertTrue(enabled_service in all_services)

    def test_start(self):
        """
        Test for Start the specified service
        """
        mock = MagicMock(return_value=True)
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
            self.assertFalse(gentoo_service.start("name"))
        mock.assert_called_once_with(
            "/etc/init.d/name start", ignore_retcode=False, python_shell=False
        )

    def test_stop(self):
        """
        Test for Stop the specified service
        """
        mock = MagicMock(return_value=True)
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
            self.assertFalse(gentoo_service.stop("name"))
        mock.assert_called_once_with(
            "/etc/init.d/name stop", ignore_retcode=False, python_shell=False
        )

    def test_restart(self):
        """
        Test for Restart the named service
        """
        mock = MagicMock(return_value=True)
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
            self.assertFalse(gentoo_service.restart("name"))
        mock.assert_called_once_with(
            "/etc/init.d/name restart", ignore_retcode=False, python_shell=False
        )

    def test_reload_(self):
        """
        Test for Reload the named service
        """
        mock = MagicMock(return_value=True)
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
            self.assertFalse(gentoo_service.reload_("name"))
        mock.assert_called_once_with(
            "/etc/init.d/name reload", ignore_retcode=False, python_shell=False
        )

    def test_zap(self):
        """
        Test for Reload the named service
        """
        mock = MagicMock(return_value=True)
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
            self.assertFalse(gentoo_service.zap("name"))
        mock.assert_called_once_with(
            "/etc/init.d/name zap", ignore_retcode=False, python_shell=False
        )

    def test_status(self):
        """
        Test for Return the status for a service
        """
        mock = MagicMock(return_value=True)
        with patch.dict(gentoo_service.__salt__, {"status.pid": mock}):
            self.assertTrue(gentoo_service.status("name", 1))

        # service is running
        mock = MagicMock(return_value=0)
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
            self.assertTrue(gentoo_service.status("name"))
        mock.assert_called_once_with(
            "/etc/init.d/name status", ignore_retcode=True, python_shell=False
        )

        # service is not running
        mock = MagicMock(return_value=1)
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
            self.assertFalse(gentoo_service.status("name"))
        mock.assert_called_once_with(
            "/etc/init.d/name status", ignore_retcode=True, python_shell=False
        )

        # service is stopped
        mock = MagicMock(return_value=3)
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
            self.assertFalse(gentoo_service.status("name"))
        mock.assert_called_once_with(
            "/etc/init.d/name status", ignore_retcode=True, python_shell=False
        )

        # service has crashed
        mock = MagicMock(return_value=32)
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": mock}):
            self.assertFalse(gentoo_service.status("name"))
        mock.assert_called_once_with(
            "/etc/init.d/name status", ignore_retcode=True, python_shell=False
        )

    def test_enable(self):
        """
        Test for Enable the named service to start at boot
        """
        rc_update_mock = MagicMock(return_value=0)
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            self.assertTrue(gentoo_service.enable("name"))
        rc_update_mock.assert_called_once_with(
            "rc-update add name", ignore_retcode=False, python_shell=False
        )
        rc_update_mock.reset_mock()

        # move service from 'l1' to 'l2' runlevel
        service_name = "name"
        runlevels = ["l1"]
        level_list_mock = MagicMock(
            return_value=self.__services({service_name: runlevels})
        )
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertTrue(gentoo_service.enable("name", runlevels="l2"))
        rc_update_mock.assert_has_calls(
            [
                call(
                    "rc-update delete name l1", ignore_retcode=False, python_shell=False
                ),
                call("rc-update add name l2", ignore_retcode=False, python_shell=False),
            ]
        )
        rc_update_mock.reset_mock()

        # requested levels are the same as the current ones
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertTrue(gentoo_service.enable("name", runlevels="l1"))
        self.assertTrue(rc_update_mock.call_count == 0)
        rc_update_mock.reset_mock()

        # same as above with the list instead of the string
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertTrue(gentoo_service.enable("name", runlevels=["l1"]))
        self.assertTrue(rc_update_mock.call_count == 0)
        rc_update_mock.reset_mock()

        # add service to 'l2' runlevel
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertTrue(gentoo_service.enable("name", runlevels=["l2", "l1"]))
        rc_update_mock.assert_called_once_with(
            "rc-update add name l2", ignore_retcode=False, python_shell=False
        )
        rc_update_mock.reset_mock()

        # remove service from 'l1' runlevel
        runlevels = ["l1", "l2"]
        level_list_mock = MagicMock(
            return_value=self.__services({service_name: runlevels})
        )
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertTrue(gentoo_service.enable("name", runlevels=["l2"]))
        rc_update_mock.assert_called_once_with(
            "rc-update delete name l1", ignore_retcode=False, python_shell=False
        )
        rc_update_mock.reset_mock()

        # move service from 'l2' add to 'l3', leaving at l1
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertTrue(gentoo_service.enable("name", runlevels=["l1", "l3"]))
        rc_update_mock.assert_has_calls(
            [
                call(
                    "rc-update delete name l2", ignore_retcode=False, python_shell=False
                ),
                call("rc-update add name l3", ignore_retcode=False, python_shell=False),
            ]
        )
        rc_update_mock.reset_mock()

        # remove from l1, l3, and add to l2, l4, and leave at l5
        runlevels = ["l1", "l3", "l5"]
        level_list_mock = MagicMock(
            return_value=self.__services({service_name: runlevels})
        )
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertTrue(
                    gentoo_service.enable("name", runlevels=["l2", "l4", "l5"])
                )
        rc_update_mock.assert_has_calls(
            [
                call(
                    "rc-update delete name l1 l3",
                    ignore_retcode=False,
                    python_shell=False,
                ),
                call(
                    "rc-update add name l2 l4", ignore_retcode=False, python_shell=False
                ),
            ]
        )
        rc_update_mock.reset_mock()

        # rc-update failed
        rc_update_mock = MagicMock(return_value=1)
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            self.assertFalse(gentoo_service.enable("name"))
        rc_update_mock.assert_called_once_with(
            "rc-update add name", ignore_retcode=False, python_shell=False
        )
        rc_update_mock.reset_mock()

        # move service delete failed
        runlevels = ["l1"]
        level_list_mock = MagicMock(
            return_value=self.__services({service_name: runlevels})
        )
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertFalse(gentoo_service.enable("name", runlevels="l2"))
        rc_update_mock.assert_called_once_with(
            "rc-update delete name l1", ignore_retcode=False, python_shell=False
        )
        rc_update_mock.reset_mock()

        # move service delete succeeds. add fails
        rc_update_mock = MagicMock()
        rc_update_mock.side_effect = [0, 1]
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertFalse(gentoo_service.enable("name", runlevels="l2"))
        rc_update_mock.assert_has_calls(
            [
                call(
                    "rc-update delete name l1", ignore_retcode=False, python_shell=False
                ),
                call("rc-update add name l2", ignore_retcode=False, python_shell=False),
            ]
        )
        rc_update_mock.reset_mock()

    def test_disable(self):
        """
        Test for Disable the named service to start at boot
        """
        rc_update_mock = MagicMock(return_value=0)
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            self.assertTrue(gentoo_service.disable("name"))
        rc_update_mock.assert_called_once_with(
            "rc-update delete name", ignore_retcode=False, python_shell=False
        )
        rc_update_mock.reset_mock()

        # disable service
        service_name = "name"
        runlevels = ["l1"]
        level_list_mock = MagicMock(
            return_value=self.__services({service_name: runlevels})
        )
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertTrue(gentoo_service.disable("name", runlevels="l1"))
        rc_update_mock.assert_called_once_with(
            "rc-update delete name l1", ignore_retcode=False, python_shell=False
        )
        rc_update_mock.reset_mock()

        # same as above with list
        runlevels = ["l1"]
        level_list_mock = MagicMock(
            return_value=self.__services({service_name: runlevels})
        )
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertTrue(gentoo_service.disable("name", runlevels=["l1"]))
        rc_update_mock.assert_called_once_with(
            "rc-update delete name l1", ignore_retcode=False, python_shell=False
        )
        rc_update_mock.reset_mock()

        # remove from 'l1', and leave at 'l2'
        runlevels = ["l1", "l2"]
        level_list_mock = MagicMock(
            return_value=self.__services({service_name: runlevels})
        )
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertTrue(gentoo_service.disable("name", runlevels=["l1"]))
        rc_update_mock.assert_called_once_with(
            "rc-update delete name l1", ignore_retcode=False, python_shell=False
        )
        rc_update_mock.reset_mock()

        # remove from non-enabled level
        runlevels = ["l2"]
        level_list_mock = MagicMock(
            return_value=self.__services({service_name: runlevels})
        )
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertTrue(gentoo_service.disable("name", runlevels=["l1"]))
        self.assertTrue(rc_update_mock.call_count == 0)
        rc_update_mock.reset_mock()

        # remove from 'l1' and 'l3', leave at 'l2'
        runlevels = ["l1", "l2", "l3"]
        level_list_mock = MagicMock(
            return_value=self.__services({service_name: runlevels})
        )
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertTrue(gentoo_service.disable("name", runlevels=["l1", "l3"]))
        rc_update_mock.assert_called_once_with(
            "rc-update delete name l1 l3", ignore_retcode=False, python_shell=False
        )
        rc_update_mock.reset_mock()

        # rc-update failed
        rc_update_mock = MagicMock(return_value=1)
        with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
            self.assertFalse(gentoo_service.disable("name"))
        rc_update_mock.assert_called_once_with(
            "rc-update delete name", ignore_retcode=False, python_shell=False
        )
        rc_update_mock.reset_mock()

        # move service delete failed
        runlevels = ["l1"]
        level_list_mock = MagicMock(
            return_value=self.__services({service_name: runlevels})
        )
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertFalse(gentoo_service.disable("name", runlevels="l1"))
        rc_update_mock.assert_called_once_with(
            "rc-update delete name l1", ignore_retcode=False, python_shell=False
        )
        rc_update_mock.reset_mock()

        # move service delete succeeds. add fails
        runlevels = ["l1", "l2", "l3"]
        level_list_mock = MagicMock(
            return_value=self.__services({service_name: runlevels})
        )
        with patch.dict(gentoo_service.__salt__, {"cmd.run": level_list_mock}):
            with patch.dict(gentoo_service.__salt__, {"cmd.retcode": rc_update_mock}):
                self.assertFalse(gentoo_service.disable("name", runlevels=["l1", "l3"]))
        rc_update_mock.assert_called_once_with(
            "rc-update delete name l1 l3", ignore_retcode=False, python_shell=False
        )
        rc_update_mock.reset_mock()

    def test_enabled(self):
        """
        Test for Return True if the named service is enabled, false otherwise
        """
        mock = MagicMock(return_value={"name": ["default"]})
        with patch.object(gentoo_service, "get_enabled", mock):
            # service is enabled at any level
            self.assertTrue(gentoo_service.enabled("name"))
            # service is enabled at the requested runlevels
            self.assertTrue(gentoo_service.enabled("name", runlevels="default"))
            # service is enabled at a different runlevels
            self.assertFalse(gentoo_service.enabled("name", runlevels="boot"))

        mock = MagicMock(return_value={"name": ["boot", "default"]})
        with patch.object(gentoo_service, "get_enabled", mock):
            # service is enabled at any level
            self.assertTrue(gentoo_service.enabled("name"))
            # service is enabled at the requested runlevels
            self.assertTrue(gentoo_service.enabled("name", runlevels="default"))
            # service is enabled at all levels
            self.assertTrue(
                gentoo_service.enabled("name", runlevels=["boot", "default"])
            )
            # service is enabled at a different runlevels
            self.assertFalse(
                gentoo_service.enabled("name", runlevels="some-other-level")
            )
            # service is enabled at a different runlevels
            self.assertFalse(
                gentoo_service.enabled("name", runlevels=["boot", "some-other-level"])
            )

    def test_disabled(self):
        """
        Test for Return True if the named service is disabled, false otherwise
        """
        mock = MagicMock(return_value=["name"])
        with patch.object(gentoo_service, "get_disabled", mock):
            self.assertTrue(gentoo_service.disabled("name"))

    def __services(self, services):
        return "\n".join(
            [" | ".join([svc, " ".join(services[svc])]) for svc in services]
        )



"""
Manage accounts in Samba's passdb using pdbedit

:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:platform:      posix

.. versionadded:: 2017.7.0
"""

import binascii
import hashlib
import logging
import re
import shlex

import salt.modules.cmdmod
import salt.utils.path

log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "pdbedit"

# Function aliases
__func_alias__ = {
    "list_users": "list",
    "get_user": "get",
}


def __virtual__():
    """
    Provides pdbedit if available
    """
    # NOTE: check for pdbedit command
    if not salt.utils.path.which("pdbedit"):
        return (False, "pdbedit command is not available")

    # NOTE: check version is >= 4.5.x
    ver = salt.modules.cmdmod.run("pdbedit -V")
    ver_regex = re.compile(r"^Version\s(\d+)\.(\d+)\.(\d+).*$")
    ver_match = ver_regex.match(ver)
    if not ver_match:
        return (False, "pdbedit -V returned an unknown version format")

    if not (int(ver_match.group(1)) >= 4 and int(ver_match.group(2)) >= 5):
        return (False, "pdbedit is to old, 4.5.0 or newer is required")

    try:
        hashlib.new("md4", "".encode("utf-16le"))
    except ValueError:
        return (False, "Hash type md4 unsupported")
    return __virtualname__


def generate_nt_hash(password):
    """
    Generate a NT HASH

    CLI Example:

    .. code-block:: bash

        salt '*' pdbedit.generate_nt_hash my_passwd
    """
    return binascii.hexlify(
        hashlib.new("md4", password.encode("utf-16le")).digest()
    ).upper()


def list_users(verbose=True, hashes=False):
    """
    List user accounts

    verbose : boolean
        return all information
    hashes : boolean
        include NT HASH and LM HASH in verbose output

    CLI Example:

    .. code-block:: bash

        salt '*' pdbedit.list
    """
    users = {} if verbose else []

    if verbose:
        # parse detailed user data
        res = __salt__["cmd.run_all"](
            "pdbedit --list --verbose {hashes}".format(
                hashes="--smbpasswd-style" if hashes else ""
            ),
        )

        if res["retcode"] > 0:
            log.error(res["stderr"] if "stderr" in res else res["stdout"])
            return users

        user_data = {}
        for user in res["stdout"].splitlines():
            if user.startswith("-"):
                if "unix username" in user_data:
                    users[user_data["unix username"]] = user_data
                user_data = {}
            elif ":" in user:
                label = user[: user.index(":")].strip().lower()
                data = user[(user.index(":") + 1) :].strip()
                user_data[label] = data

        if user_data:
            users[user_data["unix username"]] = user_data
    else:
        # list users
        res = __salt__["cmd.run_all"]("pdbedit --list")

        if res["retcode"] > 0:
            return {"Error": res["stderr"] if "stderr" in res else res["stdout"]}

        for user in res["stdout"].splitlines():
            if ":" not in user:
                continue
            user_data = user.split(":")
            if len(user_data) >= 3:
                users.append(user_data[0])

    return users


def get_user(login, hashes=False):
    """
    Get user account details

    login : string
        login name
    hashes : boolean
        include NTHASH and LMHASH in verbose output

    CLI Example:

    .. code-block:: bash

        salt '*' pdbedit.get kaylee
    """
    users = list_users(verbose=True, hashes=hashes)
    return users[login] if login in users else {}


def delete(login):
    """
    Delete user account

    login : string
        login name

    CLI Example:

    .. code-block:: bash

        salt '*' pdbedit.delete wash
    """
    if login in list_users(False):
        res = __salt__["cmd.run_all"](
            "pdbedit --delete {login}".format(login=shlex.quote(login)),
        )

        if res["retcode"] > 0:
            return {login: res["stderr"] if "stderr" in res else res["stdout"]}

        return {login: "deleted"}

    return {login: "absent"}


def create(login, password, password_hashed=False, machine_account=False):
    """
    Create user account

    login : string
        login name
    password : string
        password
    password_hashed : boolean
        set if password is a nt hash instead of plain text
    machine_account : boolean
        set to create a machine trust account instead

    CLI Example:

    .. code-block:: bash

        salt '*' pdbedit.create zoe 9764951149F84E770889011E1DC4A927 nthash
        salt '*' pdbedit.create river  1sw4ll0w3d4bug
    """
    ret = "unchanged"

    # generate nt hash if needed
    if password_hashed:
        password_hash = password.upper()
        password = ""  # wipe password
    else:
        password_hash = generate_nt_hash(password)

    # create user
    if login not in list_users(False):
        # NOTE: --create requires a password, even if blank
        res = __salt__["cmd.run_all"](
            cmd="pdbedit --create --user {login} -t {machine}".format(
                login=shlex.quote(login),
                machine="--machine" if machine_account else "",
            ),
            stdin="{password}\n{password}\n".format(password=password),
        )

        if res["retcode"] > 0:
            return {login: res["stderr"] if "stderr" in res else res["stdout"]}

        ret = "created"

    # update password if needed
    user = get_user(login, True)
    if user["nt hash"] != password_hash:
        res = __salt__["cmd.run_all"](
            "pdbedit --modify --user {login} --set-nt-hash={nthash}".format(
                login=shlex.quote(login), nthash=shlex.quote(password_hash)
            ),
        )

        if res["retcode"] > 0:
            return {login: res["stderr"] if "stderr" in res else res["stdout"]}

        if ret != "created":
            ret = "updated"

    return {login: ret}


def modify(
    login,
    password=None,
    password_hashed=False,
    domain=None,
    profile=None,
    script=None,
    drive=None,
    homedir=None,
    fullname=None,
    account_desc=None,
    account_control=None,
    machine_sid=None,
    user_sid=None,
    reset_login_hours=False,
    reset_bad_password_count=False,
):
    """
    Modify user account

    login : string
        login name
    password : string
        password
    password_hashed : boolean
        set if password is a nt hash instead of plain text
    domain : string
        users domain
    profile : string
        profile path
    script : string
        logon script
    drive : string
        home drive
    homedir : string
        home directory
    fullname : string
        full name
    account_desc : string
        account description
    machine_sid : string
        specify the machines new primary group SID or rid
    user_sid : string
        specify the users new primary group SID or rid
    account_control : string
        specify user account control properties

        .. note::
            Only the following can be set:
            - N: No password required
            - D: Account disabled
            - H: Home directory required
            - L: Automatic Locking
            - X: Password does not expire
    reset_login_hours : boolean
        reset the users allowed logon hours
    reset_bad_password_count : boolean
        reset the stored bad login counter

    .. note::
        if user is absent and password is provided, the user will be created

    CLI Example:

    .. code-block:: bash

        salt '*' pdbedit.modify inara fullname='Inara Serra'
        salt '*' pdbedit.modify simon password=r1v3r
        salt '*' pdbedit.modify jane drive='V:' homedir='\\\\serenity\\jane\\profile'
        salt '*' pdbedit.modify mal account_control=NX
    """
    ret = "unchanged"

    # flag mapping
    flags = {
        "domain": "--domain=",
        "full name": "--fullname=",
        "account desc": "--account-desc=",
        "home directory": "--homedir=",
        "homedir drive": "--drive=",
        "profile path": "--profile=",
        "logon script": "--script=",
        "account flags": "--account-control=",
        "user sid": "-U ",
        "machine sid": "-M ",
    }

    # field mapping
    provided = {
        "domain": domain,
        "full name": fullname,
        "account desc": account_desc,
        "home directory": homedir,
        "homedir drive": drive,
        "profile path": profile,
        "logon script": script,
        "account flags": account_control,
        "user sid": user_sid,
        "machine sid": machine_sid,
    }

    # update password
    if password:
        ret = create(login, password, password_hashed)[login]
        if ret not in ["updated", "created", "unchanged"]:
            return {login: ret}
    elif login not in list_users(False):
        return {login: "absent"}

    # check for changes
    current = get_user(login, hashes=True)
    changes = {}
    for key, val in provided.items():
        if key in ["user sid", "machine sid"]:
            if (
                val is not None
                and key in current
                and not current[key].endswith(str(val))
            ):
                changes[key] = str(val)
        elif key in ["account flags"]:
            if val is not None:
                if val.startswith("["):
                    val = val[1:-1]
                new = []
                for f in val.upper():
                    if f not in ["N", "D", "H", "L", "X"]:
                        log.warning(
                            "pdbedit.modify - unknown %s flag for account_control, ignored",
                            f,
                        )
                    else:
                        new.append(f)
                changes[key] = "[{flags}]".format(flags="".join(new))
        else:
            if val is not None and key in current and current[key] != val:
                changes[key] = val

    # apply changes
    if len(changes) > 0 or reset_login_hours or reset_bad_password_count:
        cmds = []
        for change in changes:
            cmds.append(
                "{flag}{value}".format(
                    flag=flags[change],
                    value=shlex.quote(changes[change]),
                )
            )
        if reset_login_hours:
            cmds.append("--logon-hours-reset")
        if reset_bad_password_count:
            cmds.append("--bad-password-count-reset")

        res = __salt__["cmd.run_all"](
            "pdbedit --modify --user {login} {changes}".format(
                login=shlex.quote(login),
                changes=" ".join(cmds),
            ),
        )

        if res["retcode"] > 0:
            return {login: res["stderr"] if "stderr" in res else res["stdout"]}

        if ret != "created":
            ret = "updated"

    return {login: ret}


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

from salt.cloud.clouds import opennebula
from salt.exceptions import SaltCloudNotFound, SaltCloudSystemExit
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

try:
    from lxml import etree  # pylint: disable=unused-import

    HAS_XML_LIBS = True
except ImportError:
    HAS_XML_LIBS = False

VM_NAME = "my-vm"


class OpenNebulaTestCase(TestCase, LoaderModuleMockMixin):
    """
    Unit TestCase for salt.cloud.clouds.opennebula module.
    """

    def setup_loader_modules(self):
        return {
            opennebula: {
                "__utils__": {"cloud.cache_node": MagicMock()},
                "__active_provider_name__": "",
            }
        }

    def test_avail_images_action(self):
        """
        Tests that a SaltCloudSystemExit error is raised when trying to call
        avail_images with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.avail_images, "action")

    def test_avail_locations_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call avail_locations
        with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.avail_locations, "action")

    def test_avail_sizes_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call avail_sizes
        with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.avail_sizes, "action")

    def test_avail_sizes(self):
        """
        Tests that avail_sizes returns an empty dictionary.
        """
        self.assertEqual(opennebula.avail_sizes(call="foo"), {})

    def test_list_clusters_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call list_clusters
        with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.list_clusters, "action")

    def test_list_datastores_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call list_datastores
        with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.list_datastores, "action")

    def test_list_hosts_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call list_datastores
        with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.list_hosts, "action")

    def test_list_nodes_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call list_nodes
        with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.list_nodes, "action")

    def test_list_nodes_full_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call list_nodes_full
        with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.list_nodes_full, "action")

    def test_list_nodes_select_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call list_nodes_full
        with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.list_nodes_select, "action")

    def test_list_security_groups_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call
        list_security_groups with --action or -a.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.list_security_groups, "action"
        )

    def test_list_templates_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call list_templates
        with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.list_templates, "action")

    def test_list_vns_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call list_vns
        with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.list_vns, "action")

    def test_reboot_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call reboot
        with anything other that --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.reboot, "my-vm", "foo")

    def test_start_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call start
        with anything other that --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.start, "my-vm", "foo")

    def test_stop_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call stop
        with anything other that --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.stop, "my-vm", "foo")

    def test_get_cluster_id_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call
        get_cluster_id with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.get_cluster_id, call="action")

    def test_get_cluster_id_no_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.get_cluster_id, None, call="foo"
        )

    def test_get_cluster_id_not_found(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        with patch(
            "salt.cloud.clouds.opennebula.list_clusters",
            MagicMock(return_value={"foo": {"id": "bar"}}),
        ):
            self.assertRaises(
                SaltCloudSystemExit,
                opennebula.get_cluster_id,
                kwargs={"name": "test"},
                call="function",
            )

    def test_get_cluster_id_success(self):
        """
        Tests that the function returns successfully.
        """
        with patch(
            "salt.cloud.clouds.opennebula.list_clusters",
            MagicMock(return_value={"test-cluster": {"id": "100"}}),
        ):
            mock_id = "100"
            mock_kwargs = {"name": "test-cluster"}
            self.assertEqual(opennebula.get_cluster_id(mock_kwargs, "foo"), mock_id)

    def test_get_datastore_id_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call
        get_datastore_id with --action or -a.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.get_datastore_id, call="action"
        )

    def test_get_datastore_id_no_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.get_datastore_id, None, call="foo"
        )

    def test_get_datastore_id_not_found(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        with patch(
            "salt.cloud.clouds.opennebula.list_datastores",
            MagicMock(return_value={"test-datastore": {"id": "100"}}),
        ):
            self.assertRaises(
                SaltCloudSystemExit,
                opennebula.get_datastore_id,
                kwargs={"name": "test"},
                call="function",
            )

    def test_get_datastore_id_success(self):
        """
        Tests that the function returns successfully.
        """
        with patch(
            "salt.cloud.clouds.opennebula.list_datastores",
            MagicMock(return_value={"test-datastore": {"id": "100"}}),
        ):
            mock_id = "100"
            mock_kwargs = {"name": "test-datastore"}
            self.assertEqual(opennebula.get_datastore_id(mock_kwargs, "foo"), mock_id)

    def test_get_host_id_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call
        get_host_id with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.get_host_id, call="action")

    def test_get_host_id_no_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.get_host_id, None, call="foo")

    def test_get_host_id_not_found(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        with patch(
            "salt.cloud.clouds.opennebula.avail_locations",
            MagicMock(return_value={"test-host": {"id": "100"}}),
        ):
            self.assertRaises(
                SaltCloudSystemExit,
                opennebula.get_host_id,
                kwargs={"name": "test"},
                call="function",
            )

    def test_get_host_id_success(self):
        """
        Tests that the function returns successfully.
        """
        with patch(
            "salt.cloud.clouds.opennebula.avail_locations",
            MagicMock(return_value={"test-host": {"id": "100"}}),
        ):
            mock_id = "100"
            mock_kwargs = {"name": "test-host"}
            self.assertEqual(opennebula.get_host_id(mock_kwargs, "foo"), mock_id)

    def test_get_image_not_found(self):
        """
        Tests that a SaltCloudNotFound is raised when the image doesn't exist.
        """
        with patch(
            "salt.cloud.clouds.opennebula.avail_images", MagicMock(return_value={})
        ):
            with patch(
                "salt.config.get_cloud_config_value", MagicMock(return_value="foo")
            ):
                self.assertRaises(SaltCloudNotFound, opennebula.get_image, "my-vm")

    def test_get_image_success(self):
        """
        Tests that the image is returned successfully.
        """
        with patch(
            "salt.cloud.clouds.opennebula.avail_images",
            MagicMock(return_value={"my-vm": {"name": "my-vm", "id": 0}}),
        ):
            with patch(
                "salt.config.get_cloud_config_value", MagicMock(return_value="my-vm")
            ):
                self.assertEqual(opennebula.get_image("my-vm"), 0)

    def test_get_image_id_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call
        get_image_id with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.get_image_id, call="action")

    def test_get_image_id_no_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.get_image_id, None, call="foo"
        )

    def test_get_image_id_not_found(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        with patch(
            "salt.cloud.clouds.opennebula.avail_images",
            MagicMock(return_value={"test-image": {"id": "100"}}),
        ):
            self.assertRaises(
                SaltCloudSystemExit,
                opennebula.get_image_id,
                kwargs={"name": "test"},
                call="function",
            )

    def test_get_image_id_success(self):
        """
        Tests that the function returns successfully.
        """
        with patch(
            "salt.cloud.clouds.opennebula.avail_images",
            MagicMock(return_value={"test-image": {"id": "100"}}),
        ):
            mock_id = "100"
            mock_kwargs = {"name": "test-image"}
            self.assertEqual(opennebula.get_image_id(mock_kwargs, "foo"), mock_id)

    def test_get_location_not_found(self):
        """
        Tests that a SaltCloudNotFound is raised when the location doesn't exist.
        """
        with patch(
            "salt.cloud.clouds.opennebula.avail_locations", MagicMock(return_value={})
        ):
            with patch(
                "salt.config.get_cloud_config_value", MagicMock(return_value="foo")
            ):
                self.assertRaises(SaltCloudNotFound, opennebula.get_location, "my-vm")

    def test_get_location_success(self):
        """
        Tests that the image is returned successfully.
        """
        with patch(
            "salt.cloud.clouds.opennebula.avail_locations",
            MagicMock(return_value={"my-host": {"name": "my-host", "id": 0}}),
        ):
            with patch(
                "salt.config.get_cloud_config_value", MagicMock(return_value="my-host")
            ):
                self.assertEqual(opennebula.get_location("my-host"), 0)

    def test_get_secgroup_id_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call
        get_host_id with --action or -a.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.get_secgroup_id, call="action"
        )

    def test_get_secgroup_id_no_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.get_secgroup_id, None, call="foo"
        )

    def test_get_secgroup_id_not_found(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        with patch(
            "salt.cloud.clouds.opennebula.list_security_groups",
            MagicMock(return_value={"test-security-group": {"id": "100"}}),
        ):
            self.assertRaises(
                SaltCloudSystemExit,
                opennebula.get_secgroup_id,
                kwargs={"name": "test"},
                call="function",
            )

    def test_get_secgroup_id_success(self):
        """
        Tests that the function returns successfully.
        """
        with patch(
            "salt.cloud.clouds.opennebula.list_security_groups",
            MagicMock(return_value={"test-secgroup": {"id": "100"}}),
        ):
            mock_id = "100"
            mock_kwargs = {"name": "test-secgroup"}
            self.assertEqual(opennebula.get_secgroup_id(mock_kwargs, "foo"), mock_id)

    def test_get_template_id_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call
        get_template_id with --action or -a.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.get_template_id, call="action"
        )

    def test_get_template_id_no_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.get_template_id, None, call="foo"
        )

    def test_get_template_id_not_found(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        with patch(
            "salt.cloud.clouds.opennebula.list_templates",
            MagicMock(return_value={"test-template": {"id": "100"}}),
        ):
            self.assertRaises(
                SaltCloudSystemExit,
                opennebula.get_template_id,
                kwargs={"name": "test"},
                call="function",
            )

    def test_get_template_id_success(self):
        """
        Tests that the function returns successfully.
        """
        with patch(
            "salt.cloud.clouds.opennebula.list_templates",
            MagicMock(return_value={"test-template": {"id": "100"}}),
        ):
            mock_id = "100"
            mock_kwargs = {"name": "test-template"}
            self.assertEqual(opennebula.get_template_id(mock_kwargs, "foo"), mock_id)

    def test_get_vm_id_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call
        get_vm_id with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.get_vm_id, call="action")

    def test_get_vm_id_no_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.get_vm_id, None, call="foo")

    def test_get_vm_id_not_found(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        with patch(
            "salt.cloud.clouds.opennebula.list_nodes",
            MagicMock(return_value={"test-vm": {"id": "100"}}),
        ):
            self.assertRaises(
                SaltCloudSystemExit,
                opennebula.get_vm_id,
                kwargs={"name": "test"},
                call="function",
            )

    def test_get_vm_id_success(self):
        """
        Tests that the function returns successfully.
        """
        with patch(
            "salt.cloud.clouds.opennebula.list_nodes",
            MagicMock(return_value={"test-vm": {"id": "100"}}),
        ):
            mock_id = "100"
            mock_kwargs = {"name": "test-vm"}
            self.assertEqual(opennebula.get_vm_id(mock_kwargs, "foo"), mock_id)

    def test_get_vn_id_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when trying to call
        get_vn_id with --action or -a.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.get_vn_id, call="action")

    def test_get_vn_id_no_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.get_vn_id, None, call="foo")

    def test_get_vn_id_not_found(self):
        """
        Tests that a SaltCloudSystemExit is raised when no name is provided.
        """
        with patch(
            "salt.cloud.clouds.opennebula.list_vns",
            MagicMock(return_value={"test-vn": {"id": "100"}}),
        ):
            self.assertRaises(
                SaltCloudSystemExit,
                opennebula.get_vn_id,
                kwargs={"name": "test"},
                call="function",
            )

    def test_get_vn_id_success(self):
        """
        Tests that the function returns successfully.
        """
        with patch(
            "salt.cloud.clouds.opennebula.list_vns",
            MagicMock(return_value={"test-vn": {"id": "100"}}),
        ):
            mock_id = "100"
            mock_kwargs = {"name": "test-vn"}
            self.assertEqual(opennebula.get_vn_id(mock_kwargs, "foo"), mock_id)

    # TODO: Write tests for create function

    def test_destroy_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.destroy, "my-vm", "function")

    def test_image_allocate_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.image_allocate, "foo")

    def test_image_allocate_no_name_or_datastore_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when a neither a datastore_id
        nor a datastore_name is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.image_allocate, "function")

    def test_image_allocate_no_path_or_data(self):
        """
        Tests that a SaltCloudSystemExit is raised when neither the path nor data args
        are provided.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.image_allocate,
            "function",
            kwargs={"datastore_id": "5"},
        )

    def test_image_clone_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.image_clone, "foo")

    def test_image_clone_no_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when a name isn't provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.image_clone, "function")

    def test_image_clone_no_image_id_or_image_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when neither the image_id nor
        the image_name args are provided.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.image_clone,
            "function",
            kwargs={"name": "test"},
        )

    @skipIf(True, "Need to figure out how to mock calls to the O.N. API first.")
    def test_image_clone_success(self):
        """
        Tests that image_clone returns successfully
        """
        with patch("image.clone", MagicMock(return_value=[True, 11, 0])):
            name = "test-image"
            expected = {
                "action": "image.clone",
                "cloned": "True",
                "cloned_image_id": "11",
                "cloned_image_name": name,
                "error_code": "0",
            }
            ret = opennebula.image_clone(
                "function", kwargs={"name": name, "image_id": 1}
            )
            self.assertEqual(expected, ret)

    def test_image_delete_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.image_delete, "foo")

    def test_image_delete_no_name_or_image_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when a neither an image_id
        nor a name is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.image_delete, "function")

    def test_image_info_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.image_info, "foo")

    def test_image_info_no_image_id_or_image_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when a neither an image_id
        nor a name is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.image_info, "function")

    def test_image_persist_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.image_persistent, "foo")

    def test_image_persist_no_persist(self):
        """
        Tests that a SaltCloudSystemExit is raised when the persist kwarg is missing.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.image_persistent, "function")

    def test_image_persist_no_name_or_image_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when a neither an image_id
        nor a name is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.image_delete,
            "function",
            kwargs={"persist": False},
        )

    def test_image_snapshot_delete_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.image_snapshot_delete, call="foo"
        )

    def test_image_snapshot_delete_no_snapshot_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the snapshot_id kwarg is
        missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.image_snapshot_delete,
            call="function",
            kwargs=None,
        )

    def test_image_snapshot_delete_no_image_name_or_image_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the image_id and image_name
        kwargs are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.image_snapshot_delete,
            call="function",
            kwargs={"snapshot_id": 0},
        )

    def test_image_snapshot_revert_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.image_snapshot_revert, call="foo"
        )

    def test_image_snapshot_revert_no_snapshot_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the snapshot_id kwarg is
        missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.image_snapshot_revert,
            call="function",
            kwargs=None,
        )

    def test_image_snapshot_revert_no_image_name_or_image_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the image_id and image_name
        kwargs are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.image_snapshot_revert,
            call="function",
            kwargs={"snapshot_id": 0},
        )

    def test_image_snapshot_flatten_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.image_snapshot_flatten, call="foo"
        )

    def test_image_snapshot_flatten_no_snapshot_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the snapshot_id kwarg is
        missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.image_snapshot_flatten,
            call="function",
            kwargs=None,
        )

    def test_image_snapshot_flatten_no_image_name_or_image_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the image_id and image_name
        kwargs are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.image_snapshot_flatten,
            call="function",
            kwargs={"snapshot_id": 0},
        )

    def test_image_update_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.image_update, "foo")

    def test_image_update_no_update_type(self):
        """
        Tests that a SaltCloudSystemExit is raised when the update_type kwarg is
        missing.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.image_update, "function")

    def test_image_update_bad_update_type_value(self):
        """
        Tests that a SaltCloudSystemExit is raised when the update_type kwarg is
        not a valid value.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.image_update,
            "function",
            kwargs={"update_type": "foo"},
        )

    def test_image_update_no_image_id_or_image_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the image_id and image_name
        kwargs are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.image_update,
            "function",
            kwargs={"update_type": "merge"},
        )

    def test_image_update_no_data_or_path(self):
        """
        Tests that a SaltCloudSystemExit is raised when the data and path
        kwargs are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.image_update,
            "function",
            kwargs={"update_type": "merge", "image_id": "0"},
        )

    def test_show_instance_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.show_instance, VM_NAME, call="foo"
        )

    def test_show_instance_success(self):
        """
        Tests that the node was found successfully.
        """
        with patch(
            "salt.cloud.clouds.opennebula._get_node",
            MagicMock(return_value={"my-vm": {"name": "my-vm", "id": 0}}),
        ):
            ret = {"my-vm": {"name": "my-vm", "id": 0}}
            self.assertEqual(opennebula.show_instance("my-vm", call="action"), ret)

    def test_secgroup_allocate_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.secgroup_allocate, "foo")

    def test_secgroup_allocate_no_data_or_path(self):
        """
        Tests that a SaltCloudSystemExit is raised when the data and path
        kwargs are missing.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.secgroup_allocate, "function")

    def test_secgroup_clone_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.secgroup_clone, "foo")

    def test_secgroup_clone_no_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the name kwarg is
        missing.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.secgroup_clone, "function")

    def test_secgroup_clone_no_secgroup_id_or_secgroup_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the secgroup_id and
        secgroup_name kwargs are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.secgroup_clone,
            "function",
            kwargs={"name": "test"},
        )

    def test_secgroup_delete_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.secgroup_delete, "foo")

    def test_secgroup_delete_no_secgroup_id_or_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the secgroup_id and
        name kwargs are missing.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.secgroup_clone, "function")

    def test_secgroup_info_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.secgroup_info, "foo")

    def test_secgroup_info_no_secgroup_id_or_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the secgroup_id and
        name kwargs are missing.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.secgroup_info, "function")

    def test_secgroup_update_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.secgroup_update, "foo")

    def test_secgroup_update_no_update_type(self):
        """
        Tests that a SaltCloudSystemExit is raised when the update_type arg is
        missing.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.secgroup_update, "function")

    def test_secgroup_update_bad_update_type_value(self):
        """
        Tests that a SaltCloudSystemExit is raised when the update_type contains
        an invalid value.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.secgroup_update,
            "function",
            kwargs={"update_type": "foo"},
        )

    def test_secgroup_update_no_secgroup_id_or_secgroup_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the secgroup_id and
        secgroup_name kwargs are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.secgroup_update,
            "function",
            kwargs={"update_type": "merge"},
        )

    def test_secgroup_update_no_data_or_path(self):
        """
        Tests that a SaltCloudSystemExit is raised when the data and
        path kwargs are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.secgroup_update,
            "function",
            kwargs={"update_type": "merge", "secgroup_id": "0"},
        )

    def test_template_allocate_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.template_allocate, "foo")

    def test_template_allocate_no_data_or_path(self):
        """
        Tests that a SaltCloudSystemExit is raised when the data and
        path kwargs are missing.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.template_allocate, "function")

    def test_template_clone_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.template_clone, "foo")

    def test_template_clone_no_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the name arg is missing.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.template_clone, "function")

    def test_template_clone_no_template_name_or_template_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the template_name and
        template_id args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.template_clone,
            "function",
            kwargs={"name": "foo"},
        )

    def test_template_delete_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.template_delete, "foo")

    def test_template_delete_no_name_or_template_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the name and
        template_id args are missing.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.template_delete, "function")

    def test_template_instantiate_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.template_instantiate, "foo")

    def test_template_instantiate_no_vm_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the vm_name arg is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.template_instantiate, "function", None
        )

    def test_template_instantiate_no_template_id_or_template_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the template_name and
        template_id args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.template_instantiate,
            "function",
            kwargs={"vm_name": "test"},
        )

    def test_template_update_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.template_update, call="foo")

    def test_template_update_bad_update_type_value(self):
        """
        Tests that a SaltCloudSystemExit is raised when the update_type contains
        and invalid value.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.template_update,
            call="function",
            kwargs={"update_type": "foo"},
        )

    def test_template_update_no_template_id_or_template_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the template_id and the
        template_name args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.template_update,
            call="function",
            kwargs={"update_type": "merge"},
        )

    def test_template_update_no_data_or_path(self):
        """
        Tests that a SaltCloudSystemExit is raised when the data and the
        path args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.template_update,
            call="function",
            kwargs={"update_type": "merge", "template_id": "0"},
        )

    def test_vm_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_action, VM_NAME, call="foo"
        )

    def test_vm_action_no_action(self):
        """
        Tests that a SaltCloudSystemExit is raised when the action arg is missing
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_action, VM_NAME, call="action"
        )

    def test_vm_allocate_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.vm_allocate, "foo")

    def test_vm_allocate_no_data_or_path(self):
        """
        Tests that a SaltCloudSystemExit is raised when the data and
        path kwargs are missing.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.vm_allocate, "function")

    def test_vm_attach_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_attach, VM_NAME, call="foo"
        )

    def test_vm_attach_no_data_or_path(self):
        """
        Tests that a SaltCloudSystemExit is raised when the data and
        path kwargs are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_attach, VM_NAME, call="action"
        )

    def test_vm_attach_nic_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_attach_nic, VM_NAME, call="foo"
        )

    def test_vm_attach_nic_no_data_or_path(self):
        """
        Tests that a SaltCloudSystemExit is raised when the data and
        path kwargs are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_attach_nic, VM_NAME, call="action"
        )

    def test_vm_deploy_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_deploy, VM_NAME, call="foo"
        )

    def test_vm_deploy_no_host_id_or_host_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the host_id and the
        host_name args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_deploy,
            VM_NAME,
            call="action",
            kwargs=None,
        )

    def test_vm_detach_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_detach, VM_NAME, call="foo"
        )

    def test_vm_detach_no_disk_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the disk_id ar is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_detach, VM_NAME, call="action"
        )

    def test_vm_detach_nic_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_detach_nic, VM_NAME, call="foo"
        )

    def test_vm_detach_nic_no_nic_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the nic_id arg is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_detach_nic, VM_NAME, call="action"
        )

    def test_vm_disk_save_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_disk_save, VM_NAME, call="foo"
        )

    def test_vm_disk_save_no_disk_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the disk_id arg is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_disk_save,
            VM_NAME,
            call="action",
            kwargs={"image_name": "foo"},
        )

    def test_vm_disk_save_no_image_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the image_name arg is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_disk_save,
            VM_NAME,
            call="action",
            kwargs={"disk_id": "0"},
        )

    def test_vm_disk_snapshot_create_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_disk_snapshot_create, VM_NAME, call="foo"
        )

    def test_vm_disk_snapshot_create_no_disk_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the disk_id arg is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_disk_snapshot_create,
            VM_NAME,
            call="action",
            kwargs={"description": "foo"},
        )

    def test_vm_disk_snapshot_create_no_description(self):
        """
        Tests that a SaltCloudSystemExit is raised when the image_name arg is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_disk_snapshot_create,
            VM_NAME,
            call="action",
            kwargs={"disk_id": "0"},
        )

    def test_vm_disk_snapshot_delete_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_disk_snapshot_delete, VM_NAME, call="foo"
        )

    def test_vm_disk_snapshot_delete_no_disk_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the disk_id arg is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_disk_snapshot_delete,
            VM_NAME,
            call="action",
            kwargs={"snapshot_id": "0"},
        )

    def test_vm_disk_snapshot_delete_no_snapshot_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the snapshot_id arg is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_disk_snapshot_delete,
            VM_NAME,
            call="action",
            kwargs={"disk_id": "0"},
        )

    def test_vm_disk_snapshot_revert_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_disk_snapshot_revert, VM_NAME, call="foo"
        )

    def test_vm_disk_snapshot_revert_no_disk_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the disk_id arg is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_disk_snapshot_revert,
            VM_NAME,
            call="action",
            kwargs={"snapshot_id": "0"},
        )

    def test_vm_disk_snapshot_revert_no_snapshot_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the snapshot_id arg is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_disk_snapshot_revert,
            VM_NAME,
            call="action",
            kwargs={"disk_id": "0"},
        )

    def test_vm_info_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.vm_info, VM_NAME, call="foo")

    def test_vm_migrate_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_migrate, VM_NAME, call="foo"
        )

    def test_vm_migrate_no_datastore_id_or_datastore_name(self):
        """
        Tests that a SaltCLoudSystemExit is raised when the datastore_id and the
        datastore_name args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_migrate,
            VM_NAME,
            call="action",
            kwargs=None,
        )

    def test_vm_migrate_no_host_id_or_host_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the host_id and the
        host_name args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_migrate,
            VM_NAME,
            call="action",
            kwargs={"datastore_id": "0"},
        )

    def test_vm_monitoring_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_monitoring, VM_NAME, call="foo"
        )

    def test_vm_resize_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_resize, VM_NAME, call="foo"
        )

    def test_vm_resize_no_data_or_path(self):
        """
        Tests that a SaltCloudSystemExit is raised when the data and path args
        are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_resize,
            VM_NAME,
            call="action",
            kwargs=None,
        )

    def test_vm_snapshot_create_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_snapshot_create, VM_NAME, call="foo"
        )

    def test_vm_snapshot_create_no_snapshot_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the snapshot_name arg
        is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_snapshot_create,
            VM_NAME,
            call="action",
            kwargs=None,
        )

    def test_vm_snapshot_delete_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_snapshot_delete, VM_NAME, call="foo"
        )

    def test_vm_snapshot_delete_no_snapshot_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the snapshot_id arg
        is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_snapshot_delete,
            VM_NAME,
            call="action",
            kwargs=None,
        )

    def test_vm_snapshot_revert_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_snapshot_revert, VM_NAME, call="foo"
        )

    def test_vm_snapshot_revert_no_snapshot_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the snapshot_id arg
        is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_snapshot_revert,
            VM_NAME,
            call="action",
            kwargs=None,
        )

    def test_vm_update_action_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --action or -a is provided.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vm_update, VM_NAME, call="foo"
        )

    def test_vm_update_no_update_type(self):
        """
        Tests that a SaltCloudSystemExit is raised when the update_type arg
        is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_update,
            VM_NAME,
            call="action",
            kwargs=None,
        )

    def test_vm_update_bad_update_type_value(self):
        """
        Tests that a SaltCloudSystemExit is raised when the update_type kwarg is
        not a valid value.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_update,
            VM_NAME,
            call="action",
            kwargs={"update_type": "foo"},
        )

    def test_vm_update_no_data_or_path(self):
        """
        Tests that a SaltCloudSystemExit is raised when the data and path args
        are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vm_update,
            VM_NAME,
            call="action",
            kwargs={"update_type": "merge"},
        )

    def test_vn_add_ar_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.vn_add_ar, call="foo")

    def test_vn_add_ar_no_vn_id_or_vn_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the vn_id and vn_name
        args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vn_add_ar, call="function", kwargs=None
        )

    def test_vn_add_ar_no_path_or_data(self):
        """
        Tests that a SaltCloudSystemExit is raised when the path and data
        args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vn_add_ar,
            call="function",
            kwargs={"vn_id": "0"},
        )

    def test_vn_allocate_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.vn_allocate, call="foo")

    def test_vn_allocate_no_data_or_path(self):
        """
        Tests that a SaltCloudSystemExit is raised when the path and data
        args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vn_allocate, call="function", kwargs=None
        )

    def test_vn_delete_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.vn_delete, call="foo")

    def test_vn_delete_no_vn_id_or_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the vn_id and name
        args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vn_delete, call="function", kwargs=None
        )

    def test_vn_free_ar_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.vn_free_ar, call="foo")

    def test_vn_free_ar_no_ar_id(self):
        """
        Tests that a SaltCloudSystemExit is raised when the ar_id is missing.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vn_free_ar, call="function", kwargs=None
        )

    def test_vn_free_ar_no_vn_id_or_vn_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the vn_id and vn_name
        args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vn_free_ar,
            call="function",
            kwargs={"ar_id": "0"},
        )

    def test_vn_hold_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.vn_hold, call="foo")

    def test_vn_hold_no_vn_id_or_vn_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the vn_id and vn_name
        args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vn_hold, call="function", kwargs=None
        )

    def test_vn_hold_no_data_or_path(self):
        """
        Tests that a SaltCloudSystemExit is raised when the data and path
        args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vn_hold,
            call="function",
            kwargs={"vn_id": "0"},
        )

    def test_vn_info_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.vn_info, call="foo")

    def test_vn_info_no_vn_id_or_vn_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the vn_id and vn_name
        args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vn_info, call="function", kwargs=None
        )

    def test_vn_release_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.vn_release, call="foo")

    def test_vn_release_no_vn_id_or_vn_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the vn_id and vn_name
        args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vn_release, call="function", kwargs=None
        )

    def test_vn_release_no_data_or_path(self):
        """
        Tests that a SaltCloudSystemExit is raised when the data and path
        args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vn_release,
            call="function",
            kwargs={"vn_id": "0"},
        )

    def test_vn_reserve_function_error(self):
        """
        Tests that a SaltCloudSystemExit is raised when something other than
        --function or -f is provided.
        """
        self.assertRaises(SaltCloudSystemExit, opennebula.vn_reserve, call="foo")

    def test_vn_reserve_no_vn_id_or_vn_name(self):
        """
        Tests that a SaltCloudSystemExit is raised when the vn_id and vn_name
        args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit, opennebula.vn_reserve, call="function", kwargs=None
        )

    def test_vn_reserve_no_data_or_path(self):
        """
        Tests that a SaltCloudSystemExit is raised when the data and path
        args are missing.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula.vn_reserve,
            call="function",
            kwargs={"vn_id": "0"},
        )

    @skipIf(not HAS_XML_LIBS, "cannot find lxml python library")
    def test__get_xml(self):
        """
        Tests that invalid XML raises SaltCloudSystemExit.
        """
        self.assertRaises(
            SaltCloudSystemExit,
            opennebula._get_xml,
            "[VirtualMachinePoolInfo] User couldn't be authenticated, aborting call.",
        )



import logging
import random
import string

import pytest
import salt.config
import salt.loader
import salt.states.boto_cognitoidentity as boto_cognitoidentity
from salt.utils.versions import LooseVersion
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

# pylint: disable=import-error,no-name-in-module
from tests.unit.modules.test_boto_cognitoidentity import (
    BotoCognitoIdentityTestCaseMixin,
)

try:
    import boto3
    from botocore.exceptions import ClientError

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False


# pylint: enable=import-error,no-name-in-module

# the boto_cognitoidentity module relies on the connect_to_region() method
# which was added in boto 2.8.0
# https://github.com/boto/boto/commit/33ac26b416fbb48a60602542b4ce15dcc7029f12
required_boto3_version = "1.2.1"

region = "us-east-1"
access_key = "GKTADJGHEIQSXMKKRBJ08H"
secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
conn_parameters = {
    "region": region,
    "key": access_key,
    "keyid": secret_key,
    "profile": {},
}
error_message = (
    "An error occurred (101) when calling the {0} operation: Test-defined error"
)
error_content = {"Error": {"Code": 101, "Message": "Test-defined error"}}

first_pool_id = "first_pool_id"
first_pool_name = "first_pool"
second_pool_id = "second_pool_id"
second_pool_name = "second_pool"
second_pool_name_updated = "second_pool_updated"
third_pool_id = "third_pool_id"
third_pool_name = first_pool_name
default_pool_name = "default_pool_name"
default_pool_id = "default_pool_id"
default_dev_provider = "test_provider_default"

identity_pools_ret = dict(
    IdentityPools=[
        dict(IdentityPoolId=first_pool_id, IdentityPoolName=first_pool_name),
        dict(IdentityPoolId=second_pool_id, IdentityPoolName=second_pool_name),
        dict(IdentityPoolId=third_pool_id, IdentityPoolName=third_pool_name),
    ]
)

first_pool_ret = dict(
    IdentityPoolId=first_pool_id,
    IdentityPoolName=first_pool_name,
    AllowUnauthenticatedIdentities=False,
    SupportedLoginProviders={
        "accounts.google.com": "testing123",
        "api.twitter.com": "testing123",
        "graph.facebook.com": "testing123",
        "www.amazon.com": "testing123",
    },
    DeveloperProviderName="test_provider",
    OpenIdConnectProviderARNs=["some_provider_arn", "another_provider_arn"],
)

first_pool_role_ret = dict(
    IdentityPoolId=first_pool_id,
    Roles=dict(
        authenticated="first_pool_auth_role", unauthenticated="first_pool_unauth_role"
    ),
)

second_pool_ret = dict(
    IdentityPoolId=second_pool_id,
    IdentityPoolName=second_pool_name,
    AllowUnauthenticatedIdentities=False,
)

second_pool_role_ret = dict(
    IdentityPoolId=second_pool_id, Roles=dict(authenticated="second_pool_auth_role")
)

second_pool_update_ret = dict(
    IdentityPoolId=second_pool_id,
    IdentityPoolName=second_pool_name,
    AllowUnauthenticatedIdentities=True,
)

third_pool_ret = dict(
    IdentityPoolId=third_pool_id,
    IdentityPoolName=third_pool_name,
    AllowUnauthenticatedIdentities=False,
    DeveloperProviderName="test_provider2",
)

third_pool_role_ret = dict(IdentityPoolId=third_pool_id)

default_pool_ret = dict(
    IdentityPoolId=default_pool_id,
    IdentityPoolName=default_pool_name,
    AllowUnauthenticatedIdentities=False,
    DeveloperProviderName=default_dev_provider,
)

default_pool_role_ret = dict(IdentityPoolId=default_pool_id)


log = logging.getLogger(__name__)


def _has_required_boto():
    """
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    """
    if not HAS_BOTO:
        return False
    elif LooseVersion(boto3.__version__) < LooseVersion(required_boto3_version):
        return False
    else:
        return True


class BotoCognitoIdentityStateTestCaseBase(TestCase, LoaderModuleMockMixin):
    conn = None

    def setup_loader_modules(self):
        ctx = {}
        utils = salt.loader.utils(
            self.opts,
            whitelist=["boto", "boto3", "args", "systemd", "path", "platform", "reg"],
            context=ctx,
        )
        serializers = salt.loader.serializers(self.opts)
        self.funcs = funcs = salt.loader.minion_mods(
            self.opts, context=ctx, utils=utils, whitelist=["boto_cognitoidentity"]
        )
        self.salt_states = salt.loader.states(
            opts=self.opts,
            functions=funcs,
            utils=utils,
            whitelist=["boto_cognitoidentity"],
            serializers=serializers,
        )
        return {
            boto_cognitoidentity: {
                "__opts__": self.opts,
                "__salt__": funcs,
                "__utils__": utils,
                "__states__": self.salt_states,
                "__serializers__": serializers,
            }
        }

    @classmethod
    def setUpClass(cls):
        cls.opts = salt.config.DEFAULT_MINION_OPTS.copy()
        cls.opts["grains"] = salt.loader.grains(cls.opts)

    @classmethod
    def tearDownClass(cls):
        del cls.opts

    def setUp(self):
        self.addCleanup(delattr, self, "funcs")
        self.addCleanup(delattr, self, "salt_states")
        # Set up MagicMock to replace the boto3 session
        # connections keep getting cached from prior tests, can't find the
        # correct context object to clear it. So randomize the cache key, to prevent any
        # cache hits
        conn_parameters["key"] = "".join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
        )

        self.patcher = patch("boto3.session.Session")
        self.addCleanup(self.patcher.stop)
        self.addCleanup(delattr, self, "patcher")
        mock_session = self.patcher.start()

        session_instance = mock_session.return_value
        self.conn = MagicMock()
        self.addCleanup(delattr, self, "conn")
        session_instance.client.return_value = self.conn


@skipIf(HAS_BOTO is False, "The boto module must be installed.")
@skipIf(
    _has_required_boto() is False,
    "The boto3 module must be greater than or equal to version {}".format(
        required_boto3_version
    ),
)
class BotoCognitoIdentityTestCase(
    BotoCognitoIdentityStateTestCaseBase, BotoCognitoIdentityTestCaseMixin
):
    """
    TestCase for salt.states.boto_cognitoidentity state.module
    """

    def _describe_identity_pool_side_effect(self, *args, **kwargs):
        if kwargs.get("IdentityPoolId") == first_pool_id:
            return first_pool_ret
        elif kwargs.get("IdentityPoolId") == second_pool_id:
            return second_pool_ret
        elif kwargs.get("IdentityPoolId") == third_pool_id:
            return third_pool_ret
        else:
            return default_pool_ret

    def test_present_when_failing_to_describe_identity_pools(self):
        """
        Tests exceptions when describing identity pools
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.side_effect = ClientError(
            error_content, "error on describe identity pool"
        )
        result = self.salt_states["boto_cognitoidentity.pool_present"](
            name="test pool present",
            IdentityPoolName=first_pool_name,
            AuthenticatedRole="my_auth_role",
            **conn_parameters
        )
        self.assertEqual(result.get("result"), False)
        self.assertTrue("error on describe identity pool" in result.get("comment", {}))

    def test_present_when_multiple_pools_with_same_name_exist(self):
        """
        Tests present on an identity pool name where it matched
        multiple pools.  The result should fail.
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.side_effect = (
            self._describe_identity_pool_side_effect
        )
        result = self.salt_states["boto_cognitoidentity.pool_present"](
            name="test pool present",
            IdentityPoolName=first_pool_name,
            AuthenticatedRole="my_auth_role",
            **conn_parameters
        )
        self.assertEqual(result.get("result"), False)
        self.assertIn(
            "{}".format([first_pool_ret, third_pool_ret]), result.get("comment", "")
        )

    def test_present_when_failing_to_create_a_new_identity_pool(self):
        """
        Tests present on an identity pool name that doesn't exist and
        an error is thrown on creation.
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.side_effect = (
            self._describe_identity_pool_side_effect
        )
        self.conn.create_identity_pool.side_effect = ClientError(
            error_content, "error on create_identity_pool"
        )
        result = self.salt_states["boto_cognitoidentity.pool_present"](
            name="test pool present",
            IdentityPoolName=default_pool_name,
            AuthenticatedRole="my_auth_role",
            **conn_parameters
        )
        self.assertEqual(result.get("result"), False)
        self.assertTrue("error on create_identity_pool" in result.get("comment", ""))
        self.assertTrue(self.conn.update_identity_pool.call_count == 0)

    def test_present_when_failing_to_update_an_existing_identity_pool(self):
        """
        Tests present on a unique instance of identity pool having the matching
        IdentityPoolName, and an error is thrown on updating the pool properties.
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.side_effect = (
            self._describe_identity_pool_side_effect
        )
        self.conn.update_identity_pool.side_effect = ClientError(
            error_content, "error on update_identity_pool"
        )
        result = self.salt_states["boto_cognitoidentity.pool_present"](
            name="test pool present",
            IdentityPoolName=second_pool_name,
            AuthenticatedRole="my_auth_role",
            AllowUnauthenticatedIdentities=True,
            **conn_parameters
        )
        self.assertEqual(result.get("result"), False)
        self.assertTrue("error on update_identity_pool" in result.get("comment", ""))
        self.assertTrue(self.conn.create_identity_pool.call_count == 0)

    def _get_identity_pool_roles_side_effect(self, *args, **kwargs):
        if kwargs.get("IdentityPoolId") == first_pool_id:
            return first_pool_role_ret
        elif kwargs.get("IdentityPoolId") == second_pool_id:
            return second_pool_role_ret
        elif kwargs.get("IdentityPoolId") == third_pool_id:
            return third_pool_role_ret
        else:
            return default_pool_role_ret

    def test_present_when_failing_to_get_identity_pool_roles(self):
        """
        Tests present on a unique instance of identity pool having the matching
        IdentityPoolName, where update_identity_pool succeeded, but an error
        is thrown on getting the identity pool role prior to setting the roles.
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.side_effect = (
            self._describe_identity_pool_side_effect
        )
        self.conn.update_identity_pool.return_value = second_pool_update_ret
        self.conn.get_identity_pool_roles.side_effect = ClientError(
            error_content, "error on get_identity_pool_roles"
        )
        result = self.salt_states["boto_cognitoidentity.pool_present"](
            name="test pool present",
            IdentityPoolName=second_pool_name,
            AuthenticatedRole="my_auth_role",
            AllowUnauthenticatedIdentities=True,
            **conn_parameters
        )
        self.assertEqual(result.get("result"), False)
        self.assertTrue("error on get_identity_pool_roles" in result.get("comment", ""))
        self.assertTrue(self.conn.create_identity_pool.call_count == 0)
        self.assertTrue(self.conn.set_identity_pool_roles.call_count == 0)

    @pytest.mark.slow_test
    def test_present_when_failing_to_set_identity_pool_roles(self):
        """
        Tests present on a unique instance of identity pool having the matching
        IdentityPoolName, where update_identity_pool succeeded, but an error
        is thrown on setting the identity pool role.
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.side_effect = (
            self._describe_identity_pool_side_effect
        )
        self.conn.update_identity_pool.return_value = second_pool_update_ret
        self.conn.get_identity_pool_roles.return_value = second_pool_role_ret
        self.conn.set_identity_pool_roles.side_effect = ClientError(
            error_content, "error on set_identity_pool_roles"
        )
        with patch.dict(
            self.funcs,
            {
                "boto_iam.describe_role": MagicMock(
                    return_value={"arn": "my_auth_role_arn"}
                )
            },
        ):
            result = self.salt_states["boto_cognitoidentity.pool_present"](
                name="test pool present",
                IdentityPoolName=second_pool_name,
                AuthenticatedRole="my_auth_role",
                AllowUnauthenticatedIdentities=True,
                **conn_parameters
            )
            self.assertEqual(result.get("result"), False)
            self.assertTrue(
                "error on set_identity_pool_roles" in result.get("comment", "")
            )
            expected_call_args = (
                dict(
                    IdentityPoolId=second_pool_id,
                    Roles={"authenticated": "my_auth_role_arn"},
                ),
            )
            self.assertTrue(
                self.conn.set_identity_pool_roles.call_args == expected_call_args
            )

    @pytest.mark.slow_test
    def test_present_when_pool_name_does_not_exist(self):
        """
        Tests the successful case of creating a new instance, and updating its
        roles
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.create_identity_pool.side_effect = (
            self._describe_identity_pool_side_effect
        )
        self.conn.get_identity_pool_roles.return_value = default_pool_role_ret
        self.conn.set_identity_pool_roles.return_value = None
        with patch.dict(
            self.funcs,
            {
                "boto_iam.describe_role": MagicMock(
                    return_value={"arn": "my_auth_role_arn"}
                )
            },
        ):
            result = self.salt_states["boto_cognitoidentity.pool_present"](
                name="test pool present",
                IdentityPoolName=default_pool_name,
                AuthenticatedRole="my_auth_role",
                AllowUnauthenticatedIdentities=True,
                DeveloperProviderName=default_dev_provider,
                **conn_parameters
            )
            self.assertEqual(result.get("result"), True)
            expected_call_args = (
                dict(
                    AllowUnauthenticatedIdentities=True,
                    IdentityPoolName=default_pool_name,
                    DeveloperProviderName=default_dev_provider,
                    SupportedLoginProviders={},
                    OpenIdConnectProviderARNs=[],
                ),
            )
            self.assertTrue(
                self.conn.create_identity_pool.call_args == expected_call_args
            )
            expected_call_args = (
                dict(
                    IdentityPoolId=default_pool_id,
                    Roles={"authenticated": "my_auth_role_arn"},
                ),
            )
            self.assertTrue(
                self.conn.set_identity_pool_roles.call_args == expected_call_args
            )
            self.assertTrue(self.conn.update_identity_pool.call_count == 0)

    @pytest.mark.slow_test
    def test_present_when_pool_name_exists(self):
        """
        Tests the successful case of updating a single instance with matching
        IdentityPoolName and its roles.
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.side_effect = (
            self._describe_identity_pool_side_effect
        )
        self.conn.update_identity_pool.return_value = second_pool_update_ret
        self.conn.get_identity_pool_roles.return_value = second_pool_role_ret
        self.conn.set_identity_pool_roles.return_value = None
        with patch.dict(
            self.funcs,
            {
                "boto_iam.describe_role": MagicMock(
                    return_value={"arn": "my_auth_role_arn"}
                )
            },
        ):
            result = self.salt_states["boto_cognitoidentity.pool_present"](
                name="test pool present",
                IdentityPoolName=second_pool_name,
                AuthenticatedRole="my_auth_role",
                AllowUnauthenticatedIdentities=True,
                **conn_parameters
            )
            self.assertEqual(result.get("result"), True)
            expected_call_args = (
                dict(
                    AllowUnauthenticatedIdentities=True,
                    IdentityPoolId=second_pool_id,
                    IdentityPoolName=second_pool_name,
                ),
            )
            self.assertTrue(
                self.conn.update_identity_pool.call_args == expected_call_args
            )
            expected_call_args = (
                dict(
                    IdentityPoolId=second_pool_id,
                    Roles={"authenticated": "my_auth_role_arn"},
                ),
            )
            self.assertTrue(
                self.conn.set_identity_pool_roles.call_args == expected_call_args
            )
            self.assertTrue(self.conn.create_identity_pool.call_count == 0)

    def test_absent_when_pool_does_not_exist(self):
        """
        Tests absent on an identity pool that does not exist.
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        result = self.salt_states["boto_cognitoidentity.pool_absent"](
            name="test pool absent",
            IdentityPoolName="no_such_pool_name",
            RemoveAllMatched=False,
            **conn_parameters
        )
        self.assertEqual(result.get("result"), True)
        self.assertEqual(result["changes"], {})

    def test_absent_when_removeallmatched_is_false_and_multiple_pools_matched(self):
        """
        Tests absent on when RemoveAllMatched flag is false and there are multiple matches
        for the given pool name
        first_pool_name is matched to first and third pool with different id's
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.side_effect = (
            self._describe_identity_pool_side_effect
        )
        result = self.salt_states["boto_cognitoidentity.pool_absent"](
            name="test pool absent",
            IdentityPoolName=first_pool_name,
            RemoveAllMatched=False,
            **conn_parameters
        )
        self.assertEqual(result.get("result"), False)
        self.assertEqual(result["changes"], {})
        self.assertTrue(
            "{}".format([first_pool_ret, third_pool_ret]) in result.get("comment", "")
        )

    def test_absent_when_failing_to_describe_identity_pools(self):
        """
        Tests exceptions when describing identity pools
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.side_effect = ClientError(
            error_content, "error on describe identity pool"
        )
        result = self.salt_states["boto_cognitoidentity.pool_absent"](
            name="test pool absent",
            IdentityPoolName=first_pool_name,
            RemoveAllMatched=False,
            **conn_parameters
        )
        self.assertEqual(result.get("result"), False)
        self.assertTrue("error on describe identity pool" in result.get("comment", {}))

    def test_absent_when_erroring_on_delete_identity_pool(self):
        """
        Tests error due to delete_identity_pools
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.side_effect = (
            self._describe_identity_pool_side_effect
        )
        self.conn.delete_identity_pool.side_effect = ClientError(
            error_content, "error on delete identity pool"
        )
        result = self.salt_states["boto_cognitoidentity.pool_absent"](
            name="test pool absent",
            IdentityPoolName=first_pool_name,
            RemoveAllMatched=True,
            **conn_parameters
        )
        self.assertEqual(result.get("result"), False)
        self.assertEqual(result["changes"], {})
        self.assertTrue("error on delete identity pool" in result.get("comment", ""))

    def test_absent_when_a_single_pool_exists(self):
        """
        Tests absent succeeds on delete when a single pool matched and
        RemoveAllMatched is False
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.return_value = second_pool_ret
        self.conn.delete_identity_pool.return_value = None
        result = self.salt_states["boto_cognitoidentity.pool_absent"](
            name="test pool absent",
            IdentityPoolName=second_pool_name,
            RemoveAllMatched=False,
            **conn_parameters
        )
        self.assertEqual(result.get("result"), True)
        expected_changes = {
            "new": {"Identity Pool Id {}".format(second_pool_id): None},
            "old": {"Identity Pool Id {}".format(second_pool_id): second_pool_name},
        }
        self.assertEqual(result["changes"], expected_changes)

    def test_absent_when_multiple_pool_exists_and_removeallmatched_flag_is_true(self):
        """
        Tests absent succeeds on delete when a multiple pools matched and
        RemoveAllMatched is True

        first_pool_name should match to first_pool_id and third_pool_id
        """
        self.conn.list_identity_pools.return_value = identity_pools_ret
        self.conn.describe_identity_pool.side_effect = (
            self._describe_identity_pool_side_effect
        )
        self.conn.delete_identity_pool.return_value = None
        result = self.salt_states["boto_cognitoidentity.pool_absent"](
            name="test pool absent",
            IdentityPoolName=first_pool_name,
            RemoveAllMatched=True,
            **conn_parameters
        )
        self.assertEqual(result.get("result"), True)
        expected_changes = {
            "new": {
                "Identity Pool Id {}".format(first_pool_id): None,
                "Identity Pool Id {}".format(third_pool_id): None,
            },
            "old": {
                "Identity Pool Id {}".format(first_pool_id): first_pool_name,
                "Identity Pool Id {}".format(third_pool_id): third_pool_name,
            },
        }
        self.assertEqual(result["changes"], expected_changes)



"""
Provide authentication using local files

.. versionadded:: 2018.3.0

The `file` auth module allows simple authentication via local files. Different
filetypes are supported, including:

  1. Text files, with passwords in plaintext or hashed
  2. Apache-style htpasswd files
  3. Apache-style htdigest files

.. note::

    The ``python-passlib`` library is required when using a ``^filetype`` of
    ``htpasswd`` or ``htdigest``.

The simplest example is a plaintext file with usernames and passwords:

.. code-block:: yaml

    external_auth:
      file:
        ^filename: /etc/insecure-user-list.txt
        gene:
          - .*
        dean:
          - test.*

In this example the ``/etc/insecure-user-list.txt`` file would be formatted
as so:

.. code-block:: text

    dean:goneFishing
    gene:OceanMan

``^filename`` is the only required parameter. Any parameter that begins with
a ``^`` is passed directly to the underlying file authentication function
via ``kwargs``, with the leading ``^`` being stripped.

The text file option is configurable to work with legacy formats:

.. code-block:: yaml

    external_auth:
      file:
        ^filename: /etc/legacy_users.txt
        ^filetype: text
        ^hashtype: md5
        ^username_field: 2
        ^password_field: 3
        ^field_separator: '|'
        trey:
          - .*

This would authenticate users against a file of the following format:

.. code-block:: text

    46|trey|16a0034f90b06bf3c5982ed8ac41aab4
    555|mike|b6e02a4d2cb2a6ef0669e79be6fd02e4
    2001|page|14fce21db306a43d3b680da1a527847a
    8888|jon|c4e94ba906578ccf494d71f45795c6cb

.. note::

    The :py:func:`hashutil.digest <salt.modules.hashutil.digest>` execution
    function is used for comparing hashed passwords, so any algorithm
    supported by that function will work.

There is also support for Apache-style ``htpasswd`` and ``htdigest`` files:

.. code-block:: yaml

    external_auth:
      file:
        ^filename: /var/www/html/.htusers
        ^filetype: htpasswd
        cory:
          - .*

When using ``htdigest`` the ``^realm`` must be set:

.. code-block:: yaml

    external_auth:
      file:
        ^filename: /var/www/html/.htdigest
        ^filetype: htdigest
        ^realm: MySecureRealm
        cory:
          - .*

"""


import logging
import os

# Import salt utils
import salt.utils.files
import salt.utils.versions

log = logging.getLogger(__name__)

__virtualname__ = "file"


def __virtual__():
    return __virtualname__


def _get_file_auth_config():
    """
    Setup defaults and check configuration variables for auth backends
    """

    config = {
        "filetype": "text",
        "hashtype": "plaintext",
        "field_separator": ":",
        "username_field": 1,
        "password_field": 2,
    }

    for opt in __opts__["external_auth"][__virtualname__]:
        if opt.startswith("^"):
            config[opt[1:]] = __opts__["external_auth"][__virtualname__][opt]

    if "filename" not in config:
        log.error(
            "salt.auth.file: An authentication file must be specified "
            "via external_auth:file:^filename"
        )
        return False

    if not os.path.exists(config["filename"]):
        log.error(
            "salt.auth.file: The configured external_auth:file:^filename (%s)"
            "does not exist on the filesystem",
            config["filename"],
        )
        return False

    config["username_field"] = int(config["username_field"])
    config["password_field"] = int(config["password_field"])

    return config


def _text(username, password, **kwargs):
    """
    The text file function can authenticate plaintext and digest methods
    that are available in the :py:func:`hashutil.digest <salt.modules.hashutil.digest>`
    function.
    """

    filename = kwargs["filename"]
    hashtype = kwargs["hashtype"]
    field_separator = kwargs["field_separator"]
    username_field = kwargs["username_field"] - 1
    password_field = kwargs["password_field"] - 1

    with salt.utils.files.fopen(filename, "r") as pwfile:
        for line in pwfile.readlines():
            fields = line.strip().split(field_separator)

            try:
                this_username = fields[username_field]
            except IndexError:
                log.error(
                    "salt.auth.file: username field (%s) does not exist in file %s",
                    username_field,
                    filename,
                )
                return False
            try:
                this_password = fields[password_field]
            except IndexError:
                log.error(
                    "salt.auth.file: password field (%s) does not exist in file %s",
                    password_field,
                    filename,
                )
                return False

            if this_username == username:
                if hashtype == "plaintext":
                    if this_password == password:
                        return True
                else:
                    # Exceptions for unknown hash types will be raised by hashutil.digest
                    if this_password == __salt__["hashutil.digest"](password, hashtype):
                        return True

                # Short circuit if we've already found the user but the password was wrong
                return False
    return False


def _htpasswd(username, password, **kwargs):
    """
    Provide authentication via Apache-style htpasswd files
    """

    from passlib.apache import HtpasswdFile

    pwfile = HtpasswdFile(kwargs["filename"])

    # passlib below version 1.6 uses 'verify' function instead of 'check_password'
    if salt.utils.versions.version_cmp(kwargs["passlib_version"], "1.6") < 0:
        return pwfile.verify(username, password)
    else:
        return pwfile.check_password(username, password)


def _htdigest(username, password, **kwargs):
    """
    Provide authentication via Apache-style htdigest files
    """

    realm = kwargs.get("realm", None)
    if not realm:
        log.error(
            "salt.auth.file: A ^realm must be defined in "
            "external_auth:file for htdigest filetype"
        )
        return False

    from passlib.apache import HtdigestFile

    pwfile = HtdigestFile(kwargs["filename"])

    # passlib below version 1.6 uses 'verify' function instead of 'check_password'
    if salt.utils.versions.version_cmp(kwargs["passlib_version"], "1.6") < 0:
        return pwfile.verify(username, realm, password)
    else:
        return pwfile.check_password(username, realm, password)


def _htfile(username, password, **kwargs):
    """
    Gate function for _htpasswd and _htdigest authentication backends
    """

    filetype = kwargs.get("filetype", "htpasswd").lower()

    try:
        import passlib

        kwargs["passlib_version"] = passlib.__version__
    except ImportError:
        log.error(
            "salt.auth.file: The python-passlib library is required for %s filetype",
            filetype,
        )
        return False

    if filetype == "htdigest":
        return _htdigest(username, password, **kwargs)
    else:
        return _htpasswd(username, password, **kwargs)


FILETYPE_FUNCTION_MAP = {"text": _text, "htpasswd": _htfile, "htdigest": _htfile}


def auth(username, password):
    """
    File based authentication

    ^filename
        The path to the file to use for authentication.

    ^filetype
        The type of file: ``text``, ``htpasswd``, ``htdigest``.

        Default: ``text``

    ^realm
        The realm required by htdigest authentication.

    .. note::
        The following parameters are only used with the ``text`` filetype.

    ^hashtype
        The digest format of the password. Can be ``plaintext`` or any digest
        available via :py:func:`hashutil.digest <salt.modules.hashutil.digest>`.

        Default: ``plaintext``

    ^field_separator
        The character to use as a delimiter between fields in a text file.

        Default: ``:``

    ^username_field
        The numbered field in the text file that contains the username, with
        numbering beginning at 1 (one).

        Default: ``1``

    ^password_field
        The numbered field in the text file that contains the password, with
        numbering beginning at 1 (one).

        Default: ``2``
    """

    config = _get_file_auth_config()

    if not config:
        return False

    auth_function = FILETYPE_FUNCTION_MAP.get(config["filetype"], "text")

    return auth_function(username, password, **config)



"""
Unit tests for salt.utils.color.py
"""


import salt.utils.color
from tests.support.unit import TestCase


class ColorUtilsTestCase(TestCase):
    def test_get_colors(self):
        ret = salt.utils.color.get_colors()
        self.assertEqual("\x1b[0;37m", str(ret["LIGHT_GRAY"]))

        ret = salt.utils.color.get_colors(use=False)
        self.assertDictContainsSubset({"LIGHT_GRAY": ""}, ret)

        ret = salt.utils.color.get_colors(use="LIGHT_GRAY")
        # LIGHT_YELLOW now == LIGHT_GRAY
        self.assertEqual(str(ret["LIGHT_YELLOW"]), str(ret["LIGHT_GRAY"]))
